#!/usr/bin/env python
# coding: utf-8

from __future__ import print_function

from scipy.interpolate import interp1d

import itertools
import sys

import matplotlib.patheffects as PathEffects
from matplotlib.patches import Patch
from matplotlib.colors import LinearSegmentedColormap, LogNorm
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import netCDF4
import shapely
import rasterio.features

import cartopy
import cartopy.crs as ccrs
import cartopy.feature as cfeature

import amrs.vis

from smartcarb import misc
from smartcarb import config as cfg



BERLIN = 13.407696, 52.516984

# with 0.05 border on Smartcarb domain
XLIM = (-1.0554999649524688, 5.2455001413822178)
YLIM = (2.7944999933242798, 8.1954997777938843)



def update_legend(legend, new_lines=None, new_labels=None):

    if new_lines is None or new_labels is None:
        new_lines = []
        new_labels = []

    ax = legend.axes
    handles, labels = ax.get_legend_handles_labels()

    handles += new_lines
    labels += new_labels

    legend._legend_box = None
    legend._init_legend_box(handles, labels)
    legend._set_loc(legend._loc)
    legend.set_title(legend.get_title().get_text())

    legend.get_figure().canvas.draw()



def plot_satellite_wind(ax, lon, lat, uwind, vwind, length=5, binning=10,
                        color='black'):

    lon = misc.spatial_binning(lon, binning)
    lat = misc.spatial_binning(lat, binning)

    # spatial binning (m/s -> knots)
    uwind = 1.944 * misc.spatial_binning(uwind, binning)
    vwind = 1.944 * misc.spatial_binning(vwind, binning)

    ax.barbs(lon, lat, uwind, vwind, length=length, color=color,
             transform=ccrs.PlateCarree())



def plot_wind_field(ax, uwind, vwind, length=5, binning=10, color='black'):

    rlon, rlat = np.meshgrid(cfg.rlon, cfg.rlat)
    rlon = misc.spatial_binning(rlon, binning)
    rlat = misc.spatial_binning(rlat, binning)

    # spatial binning (m/s -> knots)
    uwind = 1.944 * misc.spatial_binning(uwind, binning)
    vwind = 1.944 * misc.spatial_binning(vwind, binning)


    # barb plots (TODO: add rotpole projection)
    ax.barbs(rlon, rlat, uwind, vwind, length=length, color=color)




def add_scatter(ax, lon, lat, slice=Ellipsis, marker='o', color='k', ms=1.0,
                label=None):

    x = np.array(lon)[slice]
    y = np.array(lat)[slice]

    return ax.plot(x, y, marker=marker, color=color, ms=ms, ls='',
                   transform=ccrs.PlateCarree(), label=label)




def add_curve(ax, x, y, origin, epsg=cfg.EPSG, label=None,
              color=None, ls='-', marker='', mec=None):
    """\
    Add a curve to map.
    """
    if epsg is None:
        lines = ax.plot(x, y, ls=ls, color=color, label=label)
    else:
        lon, lat = misc.xy2wgs(x, y, origin, epsg=epsg, use_xarray=False)
        lines = ax.plot(lon, lat, ls=ls, color=color, transform=ccrs.PlateCarree(),
                        label=label, marker=marker, mec=mec)

    if label is not None:
        update_legend(ax.get_legend())


def add_city_labels(ax, color='black', size='medium', suffixes=None, names=None,
                    va='flex', ha='left', fontdict=None, bbox=None, ms=None,
                    do_path_effect=False, mec=None):

    if names is None:
        names = ['Berlin'] + sorted(cfg.POINT_SOURCES.keys())
        
    print(names)

    if 'Berlin' in names:
        if isinstance(ha, str):
            align = ha
        else:
            align = ha[0]

        names.remove('Berlin')
        ax.plot(BERLIN[0], BERLIN[1], 'o', color=color, ms=ms, mec=mec,
                transform=ccrs.PlateCarree(), label='_none_', zorder=10)
        t = ax.text(BERLIN[0], BERLIN[1], ' Berlin ', va='center', ha=align,
                    fontsize=size, transform=ccrs.PlateCarree(), color=color,
                    fontdict=fontdict, bbox=bbox)
        t.set_clip_on(True)
        if do_path_effect:
            t.set_path_effects([PathEffects.withStroke(linewidth=2.5, foreground='w')])

    for i, key in enumerate(names, 1):
        lon, lat = cfg.POINT_SOURCES[key]
        
        ax.plot(lon, lat, 'o', color=color, transform=ccrs.PlateCarree(), mec=mec,
                ms=ms, label='_none_', zorder=10)

        if isinstance(ha, str):
            align = ha
        else:
            align = ha[i]
            
        if key in ['Staudinger', 'Heyden', 'Melnik']:
            align = 'left'
            
        if key == 'Pocerady':
            va = 'bottom'
        elif key == 'Prunerov':
            va = 'top'
        else:
            va = 'center'
            
        key = cfg.KEY2UNICODE[key]
        t = ax.text(lon, lat, ' %s ' % key, va=va, ha=align, fontsize=size,
                    transform=ccrs.PlateCarree(), color=color, clip_on=True,
                    fontdict=fontdict, bbox=bbox)
        t.set_clip_on(True)
        if do_path_effect:
            t.set_path_effects([PathEffects.withStroke(linewidth=2.5, foreground='w')])


def draw_smartcarb_domain(ax, color='black'):
    projection = ccrs.RotatedPole(
        pole_latitude=cfg.pollat,
        pole_longitude=cfg.pollon
    )
    lon = [cfg.startlon, cfg.startlon, cfg.stoplon, cfg.stoplon, cfg.startlon]
    lat = [cfg.startlat, cfg.stoplat, cfg.stoplat, cfg.startlat, cfg.startlat]

    ax.plot(lon, lat, '-', color=color, transform=projection)


def make_cosmo_map(rlon, rlat, field, vmin=None, vmax=None, cmap='viridis',
        border=0.00, norm='linear', label='', domain='smartcarb', fig=None,
        clct=None, xlim=None, ylim=None, edgecolor='black',
        add_colorbar=True, fig_size=None):

    domain = cfg.DOMAINS[domain]
    return amrs.vis.make_field_map(rlon, rlat, field, domain=domain, vmin=vmin,
                                   vmax=vmax, cmap=cmap, border=border, norm=norm,
                                   label=label, fig=fig, fig_size=fig_size, alpha=clct,
                                   xlim=xlim, ylim=ylim, edgecolor=edgecolor,
                                   add_colorbar=add_colorbar)



def make_level2_map(lon, lat, values, fig=None, vmin=None, vmax=None,
        domain='smartcarb', label='', alpha=1.0, cmap='viridis',
        clct=None, xlim=XLIM, ylim=YLIM, edgecolor='black',
        is_discrete=False, cb_labels=None, cb_labelsize='small',
        truncate_cmap=False, do_zoom=False, zoom_area=None, fig_size=6.97,
                    dright=0.0, bg_color='silver'):
    """\
    lon, lat :: longitude and latitude of pixel corners
    values   :: 2d field of level-2 data

    is_discrete :: bool (default False)
        if True uses colormap with discrete levels with one colour per values
        in fields
    """
    domain = cfg.DOMAINS[domain]

    return amrs.vis.make_level2_map(lon, lat, values, domain=domain, fig=fig,
                                    vmin=vmin, vmax=vmax, label=label,
                                    alpha=alpha, do_zoom=do_zoom,
                                    zoom_area=zoom_area,
                                    cmap=cmap, clct=clct, xlim=xlim, ylim=ylim,
                                    edgecolor=edgecolor, is_discrete=is_discrete,
                                    cb_labels=cb_labels, cb_labelsize=cb_labelsize,
                                    truncate_cmap=truncate_cmap,
                                    fig_size=fig_size, dright=dright,
                                    bg_color=bg_color)





def main(field, rlon, rlat, rotpole=None, vmin=None, vmax=None, cmap=None,
        label=None, edgecolor='white', border=50, ax=None, img=None,
        title=None):
    """\
    Make map. `ax` and `img` can be already drawn map of which only
    the image is replaced with a new one (assuming same vmin, vmax etc.).
    """

    if label is None:
        label = 'no title'

    dlon = (rlon[1]-rlon[0])
    left = rlon[0] - dlon / 2.0
    right = rlon[-1] + dlon / 2.0

    dlat = (rlat[1]-rlat[0])
    bottom = rlat[0] - dlat / 2.0
    top = rlat[-1] + dlat / 2.0

    make_new = (ax is None and img is None)

    if make_new:

        if rotpole is None:
            rotated_pole = ccrs.RotatedPole(
                pole_latitude=cfg.pollat,
                pole_longitude=cfg.pollon
            )
        else:
            rotated_pole = ccrs.RotatedPole(
                pole_latitude=rotpole['grid_north_pole_latitude'],
                pole_longitude=rotpole['grid_north_pole_longitude']
            )

        fig = plt.figure(figsize=(5.8, 4))
        ax = plt.subplot(111, projection=rotated_pole)

        ax.set_xlim(left + border, right - border)
        ax.set_ylim(bottom + border, top - border)

        # add coastlines and countries
        ax.coastlines(resolution='10m', color=edgecolor, linewidth=0.5)
        lines = cfeature.NaturalEarthFeature(
            category='cultural',
            name='admin_0_boundary_lines_land',
            scale='10m',
        )
        ax.add_feature(lines, edgecolor=edgecolor, facecolor='none', linewidth=0.5)
        #lines = cfeature.NaturalEarthFeature(
        #    category='cultural',
        #    name='admin_1_states_provinces_lines',
        #    scale='10m',
        #)
        #ax.add_feature(lines, edgecolor=edgecolor, facecolor='none', linewidth=0.5)

    else:
        img.remove()

    # show image
    if title is not None:
        ax.set_title(title)

    img = ax.imshow(field,
        extent=(left, right, bottom, top),
        vmin=vmin, vmax=vmax, cmap=cmap
    )

    if make_new:
        cb = plt.colorbar(img, extend='both')
        cb.set_label(label)

        plt.tight_layout()
        plt.show()


    return ax, img


if __name__ == '__main__':
    pass





